import logging
import incremental
import tzlocal
from datetime import datetime, timezone
import time
from dateutil import parser
from typing import List, Dict, Union, NoReturn, Any, Tuple
from types import MethodType
from pathlib import Path
import math
import pandas as pd
import numpy as np
import asyncio
from tabulate import tabulate
import inspect

import feedparser # RSS feed parser
from ccxt.base.exchange import Exchange as CcxtExchange
import ccxt
import ccxt.pro as ccxtpro

# https://www.analyticsvidhya.com/blog/2021/06/download-financial-dataset-using-yahoo-finance-in-python-a-complete-guide/
from yahoofinancials import YahooFinancials

# yfinance allows intervals '1m', '5m', '15m', '1h', '1d', '1wk', '1mo'. yahoofinancials not as flexible
import yfinance as yf

from siglab_py.util.retry_util import retry
from siglab_py.exchanges.futubull import Futubull
from siglab_py.exchanges.any_exchange import AnyExchange
from siglab_py.exchanges.deribit import Deribit, DeribitAsync
from siglab_py.exchanges.binance import Binance, BinanceAsync

def instantiate_exchange(
    exchange_name : str,
    api_key : Union[str, None] = None,
    secret : Union[str, None]  = None,
    passphrase : Union[str, None] = None,
    default_type : Union[str, None] = 'spot',
    default_sub_type : Union[str, None] = None,
    default_max_slippage_bps : int = 100, # This is for market orders
    rate_limit_ms : float = 100
) -> Union[AnyExchange, None]:
    exchange_name = exchange_name.lower().strip()

    # Look at ccxt exchange.describe. under 'options' \ 'defaultType' (and 'defaultSubType') for what markets the exchange support.
    # https://docs.ccxt.com/en/latest/manual.html#instantiation
    exchange_params : Dict[str, Any]= {
                        'apiKey' : api_key,
                        'secret' : secret,
                        'enableRateLimit'  : True,
                        'rateLimit' : rate_limit_ms,
                        'options' : {
                            'defaultType' : default_type,
                            'defaultSlippage' : default_max_slippage_bps
                        }
                    }
    
    if default_sub_type:
        exchange_params['defaultSubType'] = default_sub_type

    if api_key:
        exchange_params['apiKey'] = api_key
    if secret:
        exchange_params['secret'] = secret
    if passphrase:
        exchange_params['passphrase'] = passphrase

    if exchange_name=='binance':
        exchange = Binance(exchange_params)  # type: ignore
    elif exchange_name=='bybit':
        exchange = ccxt.bybit(exchange_params) # type: ignore
    elif exchange_name=='okx':
        exchange = ccxt.okx(exchange_params) # type: ignore
    elif exchange_name=='deribit':
        exchange = Deribit(exchange_params)  # type: ignore
    elif exchange_name=='hyperliquid':
        exchange = ccxt.hyperliquid(
            {
                "walletAddress" : api_key, # type: ignore
                "privateKey" : secret, 
                'enableRateLimit'  : True,
                'rateLimit' : rate_limit_ms
            }
        )
        def patch_create_order_response(
                self,            
                average_price : float,
                create_order_response : Dict,
                order_type : str = "market"
            ):
                if order_type=='market':
                    create_order_response['type'] = order_type # Hyperliquid tag market orders as limit orders (but with very wide limit prices)
                    if (
                        'average' not in create_order_response 
                        or ('average' in create_order_response and not create_order_response['average'])
                        or ('average' in create_order_response and create_order_response['average']==0)
                    ):
                        create_order_response['average'] = average_price # For market orders, they tag 'average' null

        exchange.patch_create_order_response = MethodType(
            patch_create_order_response,
            exchange
        )  
    else:
        raise ValueError(f"Unsupported exchange {exchange_name}.")

    exchange.load_markets() # type: ignore

    if not hasattr(exchange, "patch_create_order_response"):
        def default_patch_create_order_response(
            self,
            average_price : float,
            create_order_response : Dict,
            order_type : str = "market"
        ):
            pass

        exchange.patch_create_order_response = MethodType(
            default_patch_create_order_response,
            exchange
        )

    return exchange # type: ignore

async def async_instantiate_exchange(
    gateway_id : str,
    api_key : str,
    secret : str,
    passphrase : str,
    default_type : Union[str, None] = 'spot',
    default_sub_type : Union[str, None] = None,
    default_max_slippage_bps : int = 100, # This is for market orders
    rate_limit_ms : float = 100,
    verbose : bool = False
) -> Union[AnyExchange, None]:
    exchange : Union[AnyExchange, None] = None
    exchange_name : str = gateway_id.split('_')[0]
    exchange_name =exchange_name.lower().strip()

    # Look at ccxt exchange.describe. under 'options' \ 'defaultType' (and 'defaultSubType') for what markets the exchange support.
    # https://docs.ccxt.com/en/latest/manual.html#instantiation
    exchange_params : Dict[str, Any]= {
                        'apiKey' : api_key,
                        'secret' : secret,
                        'enableRateLimit'  : True,
                        'rateLimit' : rate_limit_ms,
                        'options' : {
                            'defaultType' : default_type,
                            'defaultSlippage' : default_max_slippage_bps
                        },
                        'verbose': verbose
                    }
    
    if default_sub_type:
        exchange_params['defaultSubType'] = default_sub_type

    if exchange_name=='binance':
        # spot, future, margin, delivery, option
        # https://github.com/ccxt/ccxt/blob/master/python/ccxt/binance.py#L1298
        exchange = BinanceAsync(exchange_params)  # type: ignore
    elif exchange_name=='bybit':
        # spot, linear, inverse, futures
        # https://github.com/ccxt/ccxt/blob/master/python/ccxt/bybit.py#L1041
        exchange = ccxtpro.bybit(exchange_params) # type: ignore
    elif exchange_name=='okx':
        # 'funding', spot, margin, future, swap, option
        # https://github.com/ccxt/ccxt/blob/master/python/ccxt/okx.py#L1144
        exchange_params['password'] = passphrase
        exchange = ccxtpro.okx(exchange_params) # type: ignore
    elif exchange_name=='deribit':
        # spot, swap, future
        # https://github.com/ccxt/ccxt/blob/master/python/ccxt/deribit.py#L360
        exchange = DeribitAsync(exchange_params)  # type: ignore
    elif exchange_name=='kraken':
        exchange = ccxtpro.kraken(exchange_params) # type: ignore
    elif exchange_name=='hyperliquid':
        '''
        https://app.hyperliquid.xyz/API
        
        defaultType from ccxt: swap
            https://github.com/ccxt/ccxt/blob/master/python/ccxt/hyperliquid.py#L225
        
        How to integrate? You can skip first 6 min: https://www.youtube.com/watch?v=UuBr331wxr4&t=363s

        Example, 
            API credentials created under "\ More \ API":
                    Ledger Arbitrum Wallet Address: 0xAAAAA <-- This is your Ledger Arbitrum wallet address with which you connect to Hyperliquid. 
                    API Wallet Address 0xBBBBB <-- Generated
                    privateKey 0xCCCCC

        Basic connection via CCXT:
            import asyncio
            import ccxt.pro as ccxtpro

            async def main():
                rate_limit_ms = 100
                exchange_params = {
                    "walletAddress" : "0xAAAAA", # Ledger Arbitrum Wallet Address here! Not the generated address.
                    "privateKey" : "0xCCCCC"
                }
                exchange = ccxtpro.hyperliquid(exchange_params) 
                balances = await exchange.fetch_balance()
                print(balances)

            asyncio.run(main())
        '''
        exchange = ccxtpro.hyperliquid(
            {
                "walletAddress" : api_key,
                "privateKey" : secret,
                'enableRateLimit'  : True,
                'rateLimit' : rate_limit_ms,
                'verbose': verbose
            }  # type: ignore
        )
        def patch_create_order_response(
                self,
                average_price : float,
                create_order_response : Dict,
                order_type : str = "market"
            ):
                if order_type=='market':
                    create_order_response['type'] = order_type # Hyperliquid tag market orders as limit orders (but with very wide limit prices)
                    if order_type=='market':
                        create_order_response['type'] = order_type # Hyperliquid tag market orders as limit orders (but with very wide limit prices)
                        if (
                            'average' not in create_order_response 
                            or ('average' in create_order_response and not create_order_response['average'])
                            or ('average' in create_order_response and create_order_response['average']==0)
                        ):
                            create_order_response['average'] = average_price # For market orders, they tag 'average' null

        exchange.patch_create_order_response = MethodType(
            patch_create_order_response,
            exchange
        )
    else:
        raise ValueError(f"Unsupported exchange {exchange_name}, check gateway_id {gateway_id}.")

    await exchange.load_markets() # type: ignore

    if not hasattr(exchange, "patch_create_order_response"):
        def default_patch_create_order_response(
            self,
            average_price : float,
            create_order_response : Dict,
            order_type : str = "market"
        ):
            pass

        exchange.patch_create_order_response = MethodType(
            default_patch_create_order_response,
            exchange
        )

    '''
    Is this necessary? The added trouble is for example bybit.authenticate requires arg 'url'. binance doesn't. And fetch_balance already test credentials.

    try:
        await exchange.authenticate() # type: ignore
    except Exception as swallow_this_error:
        pass
    '''

    return exchange

def timestamp_to_datetime_cols(pd_candles : pd.DataFrame):
    def _fix_timestamp_ms(x):
        if isinstance(x, pd.Timestamp):
            return int(x.value // 10**6)
        elif isinstance(x, np.datetime64):
            return int(x.astype('int64') // 10**6)
        elif isinstance(x, (int, float)):
            x = int(x)
            if len(str(abs(x))) == 13:
                return x
            else:
                return int(x * 1000)
        else:
            raise ValueError(f"Unsupported type {type(x)} for timestamp conversion")
    pd_candles['timestamp_ms'] = pd_candles['timestamp_ms'].apply(_fix_timestamp_ms)
    pd_candles['datetime'] = pd_candles['timestamp_ms'].apply(lambda x: datetime.fromtimestamp(int(x/1000)))
    pd_candles['datetime'] = pd.to_datetime(pd_candles['datetime'])
    pd_candles['datetime'] = pd_candles['datetime'].dt.tz_localize(None)  # type: ignore
    pd_candles['datetime_utc'] = pd_candles['timestamp_ms'].apply(
        lambda x: datetime.fromtimestamp(int(x.timestamp()) if isinstance(x, pd.Timestamp) else int(x / 1000), tz=timezone.utc)
    )
    
    # This is to make it easy to do grouping with Excel pivot table
    pd_candles['year'] = pd_candles['datetime'].dt.year  # type: ignore
    pd_candles['month'] = pd_candles['datetime'].dt.month  # type: ignore
    pd_candles['day'] = pd_candles['datetime'].dt.day  # type: ignore
    pd_candles['hour'] = pd_candles['datetime'].dt.hour  # type: ignore
    pd_candles['minute'] = pd_candles['datetime'].dt.minute  # type: ignore
    pd_candles['dayofweek'] = pd_candles['datetime'].dt.dayofweek  # type: ignore dayofweek: Monday is 0 and Sunday is 6

    pd_candles['week_of_month'] = pd_candles['timestamp_ms'].apply(
        lambda x: timestamp_to_week_of_month(int(x/1000))
    )

    pd_candles['apac_trading_hr'] = pd_candles['timestamp_ms'].apply(
        lambda x: "APAC" in timestamp_to_active_trading_regions(int(x/1000))
    )
    pd_candles['emea_trading_hr'] = pd_candles['timestamp_ms'].apply(
        lambda x: "EMEA" in timestamp_to_active_trading_regions(int(x/1000))
    )
    pd_candles['amer_trading_hr'] = pd_candles['timestamp_ms'].apply(
        lambda x: "AMER" in timestamp_to_active_trading_regions(int(x/1000))
    )

    pd_candles['timestamp_ms_gap'] = pd_candles['timestamp_ms'] - pd_candles['timestamp_ms'].shift(1)
    
    # Depending on asset, minutes bar may have gaps
    timestamp_ms_gap_median = pd_candles['timestamp_ms_gap'].median()
    NUM_MS_IN_1HR = 60*60*1000
    if timestamp_ms_gap_median>=NUM_MS_IN_1HR:
        num_rows_with_expected_gap = pd_candles[~pd_candles.timestamp_ms_gap.isna()][pd_candles.timestamp_ms_gap==timestamp_ms_gap_median].shape[0]
        assert(num_rows_with_expected_gap/pd_candles.shape[0]>0.9)
    pd_candles.drop(columns=['timestamp_ms_gap'], inplace=True)

'''
APAC (Asia-Pacific) Trading Hours
    UTC 21:00 - 09:00 (approximate range)
    Major financial centers: Tokyo, Hong Kong, Singapore, Sydney

EMEA (Europe, Middle East, Africa) Trading Hours
    UTC 07:00 - 16:00 (approximate range)
    Major financial centers: London, Frankfurt, Paris, Zurich, Dubai

US Trading Hours
    UTC 13:00 - 22:00 (approximate range)
    Major financial centers: New York, Chicago
    Key markets: NYSE, NASDAQ

utcnow and utcfromtimestamp been deprecated in Python 3.12 
https://www.pythonmorsels.com/converting-to-utc-time/

Example, UTC 23:00 is 3rd hour in APAC trading session
    utc_hour = 23
    i = get_regions_trading_utc_hours()['APAC'].index(utc_hour)
    assert(i==2)
'''
def get_regions_trading_utc_hours():
    return {
        'APAC' : [21,22,23,0,1,2,3,4,5,6,7,8,9],
        'EMEA' : [7,8,9,10,11,12,13,14,15,16],
        'AMER' : [13,14,15,16,17,18,19,20,21,22]
    }
			
def timestamp_to_active_trading_regions(
        timestamp_ms : int
) -> List[str]:
    active_trading_regions : List[str] = []

    dt_utc = datetime.fromtimestamp(int(timestamp_ms / 1000), tz=timezone.utc)
    utc_hour = dt_utc.hour
    if utc_hour in get_regions_trading_utc_hours()['APAC']:
        active_trading_regions.append("APAC") 

    if utc_hour in get_regions_trading_utc_hours()['EMEA']:
        active_trading_regions.append("EMEA")

    if utc_hour in get_regions_trading_utc_hours()['AMER']:
        active_trading_regions.append("AMER")

    return active_trading_regions

def timestamp_to_week_of_month(timestamp_ms: int) -> int:
    """
    Returns:
        int: Week of the month (0 = first week, 1 = second week, etc.).
    """
    dt = datetime.fromtimestamp(timestamp_ms / 1000)
    day_of_month = dt.day
    week_of_month = (day_of_month - 1) // 7
    return week_of_month

def fix_column_types(pd_candles : pd.DataFrame):
    pd_candles['open'] = pd_candles['open'].astype(float)
    pd_candles['high'] = pd_candles['high'].astype(float)
    pd_candles['low'] = pd_candles['low'].astype(float)
    pd_candles['close'] = pd_candles['close'].astype(float)
    pd_candles['volume'] = pd_candles['volume'].astype(float)

    timestamp_to_datetime_cols(pd_candles)

    '''
    The 'Unnamed: 0', 'Unnamed : 1'... etc columns often appears in a DataFrame when it is saved to a file (e.g., CSV or Excel) and later loaded. 
    This usually happens if the DataFrame's index was saved along with the data, and then pandas automatically treats it as a column during the file loading process.
    We want to drop them as it'd mess up idmin, idmax calls, which will take values from 'Unnamed' instead of actual pandas index.
    '''
    pd_candles.drop(pd_candles.columns[pd_candles.columns.str.contains('unnamed',case = False)],axis = 1, inplace = True)
    pd_candles.reset_index(drop=True, inplace=True)
    pd_candles.sort_values("datetime", inplace=True)

def interval_to_ms(interval : str) -> int:
    interval_ms : int = 0
    if interval=="d":
        interval_ms = 24*60*60*1000
    elif interval=="h":
        interval_ms = 60*60*1000
    elif interval=="m":
        interval_ms = 60*1000

    return interval_ms

'''
https://polygon.io/docs/stocks
'''
class PolygonMarketDataProvider:
    pass

class NASDAQExchange:
    def __init__(self, data_dir : Union[str, None]) -> None:
        if data_dir:
            self.data_dir = data_dir
        else:
            self.data_dir = Path(__file__).resolve().parents[2] / "data/nasdaq" 

    def fetch_ohlcv(
        self,
        symbol : str,
        since : int,
        timeframe : str,
        limit : int = 1
    ) -> List:
        pd_candles = self.fetch_candles(
            symbols=[symbol],
            start_ts=int(since/1000),
            end_ts=None,
            candle_size=timeframe
        )[symbol]
        if pd_candles is not None:
            return pd_candles.values.tolist()
        else:
            return []
    
    def fetch_candles(
        self,
        start_ts,
        end_ts,
        symbols,
        candle_size
    ) -> Dict[str, Union[pd.DataFrame, None]]:
        exchange_candles : Dict[str, Union[pd.DataFrame, None]] = {}

        start_date = datetime.fromtimestamp(start_ts)
        end_date = datetime.fromtimestamp(end_ts) if end_ts else None
        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d') if end_date else None
        local_tz = datetime.now().astimezone().tzinfo

        for symbol in symbols:
            # CSV from NASDAQ: https://www.nasdaq.com/market-activity/quotes/historical
            pd_daily_candles = pd.read_csv(f"{self.data_dir}\\NASDAQ_hist_{symbol.replace('^','')}.csv")
            pd_daily_candles.rename(columns={'Date' : 'datetime', 'Open': 'open', 'High': 'high', 'Low': 'low', 'Close/Last' : 'close', 'Adj Close' : 'adj_close', 'Volume' : 'volume' }, inplace=True)
            pd_daily_candles['open'] = pd_daily_candles['open'].astype(str).str.replace('$','')
            pd_daily_candles['high'] = pd_daily_candles['high'].astype(str).str.replace('$','')
            pd_daily_candles['low'] = pd_daily_candles['low'].astype(str).str.replace('$','')
            pd_daily_candles['close'] = pd_daily_candles['close'].astype(str).str.replace('$','')
            pd_daily_candles['datetime']= pd.to_datetime(pd_daily_candles['datetime'])
            pd_daily_candles['timestamp_ms'] = pd_daily_candles.datetime.values.astype(np.int64) // 10 ** 6  # type: ignore
            pd_daily_candles['symbol'] = symbol
            pd_daily_candles['exchange'] = 'nasdaq'
            fix_column_types(pd_daily_candles)

            if candle_size=="1h":
                # Fill forward (i.e. you dont actually have hourly candles)
                start = pd_daily_candles["datetime"].min().normalize()
                end = pd_daily_candles["datetime"].max().normalize() + pd.Timedelta(days=1)
                hourly_index = pd.date_range(start=start, end=end, freq="h") # FutureWarning: 'H' is deprecated and will be removed in a future version, please use 'h' instead.
                pd_hourly_candles = pd.DataFrame({"datetime": hourly_index})
                pd_hourly_candles = pd.merge_asof(
                    pd_hourly_candles.sort_values("datetime"),
                    pd_daily_candles.sort_values("datetime"),
                    on="datetime",
                    direction="backward"
                )

                # When you fill foward, a few candles before start date can have null values (open, high, low, close, volume ...)
                first_candle_dt = pd_hourly_candles[(~pd_hourly_candles.close.isna())  & (pd_hourly_candles['datetime'].dt.time == pd.Timestamp('00:00:00').time())].iloc[0]['datetime']  # type: ignore
                pd_hourly_candles = pd_hourly_candles[pd_hourly_candles.datetime>=first_candle_dt]
                exchange_candles[symbol] = pd_hourly_candles

            elif candle_size=="1d":
                exchange_candles[symbol] = pd_daily_candles

        return exchange_candles

class YahooExchange:
    def fetch_ohlcv(
        self,
        symbol : str,
        since : int,
        timeframe : str,
        limit : int = 1
    ) -> List:
        pd_candles = self.fetch_candles(
            symbols=[symbol],
            start_ts=int(since/1000),
            end_ts=None,
            candle_size=timeframe
        )[symbol]
        if pd_candles is not None:
            return pd_candles.values.tolist()
        else:
            return []

    def fetch_candles(
        self,
        start_ts,
        end_ts,
        symbols,
        candle_size
    ) -> Dict[str, Union[pd.DataFrame, None]]:
        exchange_candles : Dict[str, Union[pd.DataFrame, None]] = {}

        start_date = datetime.fromtimestamp(start_ts)
        end_date = datetime.fromtimestamp(end_ts)
        start_date_str = start_date.strftime('%Y-%m-%d')
        end_date_str = end_date.strftime('%Y-%m-%d')
        local_tz = datetime.now().astimezone().tzinfo

        for symbol in symbols:
            # From yf, "DateTime" in UTC
            # The requested range must be within the last 730 days. Otherwise API will return empty DataFrame.
            pd_candles = yf.download(tickers=symbol, start=start_date_str, end=end_date_str, interval=candle_size)
            pd_candles.reset_index(inplace=True) # type: ignore
            pd_candles.rename(columns={ 'Date' : 'datetime', 'Datetime' : 'datetime', 'Open': 'open', 'High': 'high', 'Low': 'low', 'Close' : 'close', 'Adj Close' : 'adj_close', 'Volume' : 'volume' }, inplace=True) # type: ignore
            pd_candles['datetime'] = pd.to_datetime(pd_candles['datetime']) # type: ignore
            if pd_candles['datetime'].dt.tz is None: # type: ignore
                pd_candles['datetime'] = pd.to_datetime(pd_candles['datetime']).dt.tz_localize('UTC') # type: ignore
            pd_candles['datetime'] = pd_candles['datetime'].dt.tz_convert(local_tz) # type: ignore
            pd_candles['datetime'] = pd_candles['datetime'].dt.tz_localize(None)  # type: ignore
            pd_candles['timestamp_ms'] = pd_candles.datetime.values.astype(np.int64) // 10**6 # type: ignore
            pd_candles = pd_candles.sort_values(by=['timestamp_ms'], ascending=[True]) # type: ignore
            
            fix_column_types(pd_candles)
            pd_candles['symbol'] = symbol
            pd_candles['exchange'] = 'yahoo'
            exchange_candles[symbol] = pd_candles

        return exchange_candles

def aggregate_candles(
    interval : str,
    pd_candles : pd.DataFrame
) -> pd.DataFrame:
    if interval[-1]=='m':
        # 'm' for pandas means months!
        interval = interval.replace('m','min')
    pd_candles.set_index('datetime', inplace=True)
    pd_candles_aggregated = pd_candles.resample(interval).agg({
        'exchange' : 'first',
        'symbol' : 'first',
        'timestamp_ms' : 'first',
        
        'open': 'first',
        'high': 'max',
        'low': 'min',
        'close': 'last',
        'volume': 'sum',

        'datetime_utc' : 'first',
        'year' : 'first',
        'month' : 'first',
        'day' : 'first',
        'hour' : 'first',
        'minute' : 'first',
        'dayofweek' : 'first',
        'week_of_month' : 'first',

        'apac_trading_hr' : 'first',
        'emea_trading_hr' : 'first',
        'amer_trading_hr' : 'first',

        'pct_chg_on_close' : 'sum',

    })
    pd_candles.reset_index(inplace=True)
    pd_candles_aggregated.reset_index(inplace=True)
    return pd_candles_aggregated
    
def fetch_historical_price(
    exchange,
    normalized_symbol : str,
    timestamp_ms : int,
    ref_timeframe : str = '1m'
):
    one_candle = fetch_ohlcv_one_candle(exchange=exchange, normalized_symbol=normalized_symbol, timestamp_ms=timestamp_ms, ref_timeframe=ref_timeframe)
    reference_price = abs(one_candle['close'] + one_candle['open'])/2 if one_candle else None
    return reference_price

def fetch_ohlcv_one_candle(
    exchange,
    normalized_symbol : str,
    timestamp_ms : int,
    ref_timeframe : str = '1m'
):
    candles = exchange.fetch_ohlcv(symbol=normalized_symbol, since=int(timestamp_ms), timeframe=ref_timeframe, limit=1)
    one_candle = {
            'timestamp_ms' : candles[0][0],
            'open' : candles[0][1],
            'high' : candles[0][2],
            'low' : candles[0][3],
            'close' : candles[0][4],
            'volume' : candles[0][5] 
        } if candles and len(candles)>0 else None
    
    return one_candle
    
def fetch_candles(
    start_ts, # in sec
    end_ts, # in sec
    exchange,
    normalized_symbols,
    candle_size,

    logger = None,

    num_candles_limit : int = 100,

    ticker_change_map : List[Dict[str, Union[str, int]]] = [],

    cache_dir : Union[str, None] = None,

    list_ts_field : Union[str, None] = None,

    validation_max_gaps : int = 10,
    validation_max_end_date_intervals : int = 1
) -> Dict[str, Union[pd.DataFrame, None]]:
    exchange_candles = { '' : None }
    num_intervals = int(candle_size.replace(candle_size[-1],''))

    if end_ts>datetime.now().timestamp():
        end_ts = int(datetime.now().timestamp())

    if type(exchange) is YahooExchange:
        exchange_candles = exchange.fetch_candles(
                            start_ts=start_ts,
                            end_ts=end_ts,
                            symbols=normalized_symbols,
                            candle_size=candle_size
                        )
    elif type(exchange) is NASDAQExchange:
        exchange_candles = exchange.fetch_candles(
                            start_ts=start_ts,
                            end_ts=end_ts,
                            symbols=normalized_symbols,
                            candle_size=candle_size
                        )
    elif type(exchange) is Futubull:
        exchange_candles = exchange.fetch_candles(
                            start_ts=start_ts,
                            end_ts=end_ts,
                            symbols=normalized_symbols,
                            candle_size=candle_size
                        )
        for symbol in exchange_candles:
            pd_candles = exchange_candles[symbol]
            if not pd_candles is None:
                fix_column_types(pd_candles) # You don't want to do this from Futubull as you'd need import Futubull from there: Circular references
                
    elif issubclass(exchange.__class__, CcxtExchange):
        exchange_candles = _fetch_candles_ccxt(
            start_ts=start_ts,
            end_ts=end_ts,
            exchange=exchange,
            normalized_symbols=normalized_symbols,
            candle_size=candle_size,
            num_candles_limit=num_candles_limit,
            ticker_change_map=ticker_change_map,
            logger=logger
        )
    if num_intervals!=1:
        for symbol in exchange_candles:
            if not exchange_candles[symbol] is None:
                exchange_candles[symbol] = aggregate_candles(candle_size, exchange_candles[symbol]) #  type: ignore

    # For invalid rows missing timestamps, o/h/l/c/v, fill forward close, set volume to zero.
    for symbol in exchange_candles:
        pd_candles = exchange_candles[symbol]
        
        if pd_candles is not None:
            mask_invalid_candles = pd_candles["timestamp_ms"].isna()
            if mask_invalid_candles.any():
                pd_invalid_candles = pd_candles[mask_invalid_candles]

                if logger is not None:
                    logger.warning(f"Dropping {pd_invalid_candles.shape[0]}/{pd_candles.shape[0]} rows from {symbol} candles (null timestamp_ms)") # type: ignore
                    logger.warning(f"{tabulate(pd_invalid_candles, headers='keys', tablefmt='psql')}") # type: ignore
                    
                def _to_timestamp_ms(dt):
                    if pd.isna(dt):
                        return pd.NA
                    if isinstance(dt, str):
                        dt = pd.to_datetime(dt)
                    return int(dt.timestamp() * 1000)
                
                pd_candles.loc[mask_invalid_candles, "timestamp_ms"] = pd_candles.loc[mask_invalid_candles, "datetime"].apply(_to_timestamp_ms)

                pd_candles["close"] = pd_candles["close"].ffill()
                pd_candles.loc[mask_invalid_candles, ["open", "high", "low"]] = pd_candles.loc[
                                                                                    mask_invalid_candles, ["close"]
                                                                                ]
                pd_candles.loc[mask_invalid_candles, "volume"] = 0.0

    return exchange_candles # type: ignore

'''
Find listing date https://gist.github.com/mr-easy/5185b1dcdd5f9f908ff196446f092e9b

Usage:
    listing_ts = find_start_time(exchange, 'HYPE/USDT:USDT', int(datetime(2024,1,1).timestamp()*1000), int(datetime(2025,5,1).timestamp()*1000), '1h')

Caveats: 
1) If listing date lies outside [start_time, end_time], this function will stackoverflow, 
2) Even if not, it's still very time consuming.

Alternative: market['created']
'''
def search_listing_ts(exchange, symbol, start_time, end_time, timeframe):
    mid_time = (start_time + end_time)//2
    if(mid_time == start_time): return mid_time+1
    ohlcv = exchange.fetch_ohlcv(symbol, timeframe, mid_time, limit=1)
    time.sleep(1)
    if(len(ohlcv) == 0):
        return search_listing_ts(exchange, symbol, mid_time, end_time, timeframe)
    else:
        return search_listing_ts(exchange, symbol, start_time, mid_time, timeframe)
    
def _fetch_candles_ccxt(
    start_ts : int,
    end_ts : int,
    exchange,
    normalized_symbols : List[str],
    candle_size : str,
    num_candles_limit : int = 100,
    ticker_change_map : List[Dict[str, Union[str, int]]] = [],
    logger = None
) -> Dict[str, Union[pd.DataFrame, None]]:
    rsp = {}

    exchange.load_markets()
    
    num_tickers = len(normalized_symbols)
    i = 0
    for ticker in normalized_symbols:
        old_ticker = get_old_ticker(ticker, ticker_change_map)
        ticker_change_mapping = get_ticker_map(ticker, ticker_change_map)

        @retry(num_attempts=3, pause_between_retries_ms=1000, logger=logger)
        def _fetch_ohlcv(exchange, symbol, timeframe, since, limit, params) -> Union[List, NoReturn]:
                one_timeframe = f"1{timeframe[-1]}"
                candles = exchange.fetch_ohlcv(symbol=symbol, timeframe=one_timeframe, since=since, limit=limit, params=params)
                if candles and len(candles)>0:
                    candles.sort(key=lambda x : x[0], reverse=False)

                return candles
            
        def _calc_increment(candle_size):
            increment = 1
            num_intervals = int(candle_size.replace(candle_size[-1],''))
            interval_type = candle_size[-1]
            if interval_type == "m":
                increment = 60
            elif interval_type == "h":
                increment = 60*60
            elif interval_type == "d":
                increment = 60*60*24
            else:
                raise ValueError(f"Invalid candle_size {candle_size}")
            return num_intervals * increment
        
        if logger:
            logger.info(f"{i}/{num_tickers} Fetching {candle_size} candles for {ticker}.")

        '''
        It uses a while loop to implement a sliding window to download candles between start_ts and end_ts. 
        However, start_ts for example can be 1 Jan 2021 for a given ticker. 
        But if that ticker listing date is 1 Jan 2025, this while loop would waste a lot of time loop between 1 Jan 2021 thru 31 Dec 2024, slowly incrementing this_cutoff += _calc_increment(candle_size).
        A more efficient way is to find listing date. Start looping from there.
        '''
        market = exchange.markets[ticker] if ticker in exchange.markets else None
        if not market:
            market = exchange.markets[old_ticker] if old_ticker else None
            if not market:
                raise ValueError(f"market {ticker} not support by exchange {exchange.name}!")

        this_ticker_start_ts = start_ts
        if market['created']:
            this_ticker_start_ts = max(this_ticker_start_ts, int(market['created']/1000))

        all_candles = []
        params = {}
        this_cutoff = this_ticker_start_ts
        while this_cutoff<end_ts:
            _ticker = ticker
            if ticker_change_mapping:
                ticker_change_cutoff_sec = int(ticker_change_mapping['cutoff_ms']) / 1000
                if this_cutoff<ticker_change_cutoff_sec:
                    _ticker = old_ticker
            candles = _fetch_ohlcv(exchange=exchange, symbol=_ticker, timeframe=candle_size, since=int(this_cutoff * 1000), limit=num_candles_limit, params=params)
            if candles and len(candles)>0:
                all_candles = all_candles + [[ int(x[0]), float(x[1]), float(x[2]), float(x[3]), float(x[4]), float(x[5]) ] for x in candles if x[1] and x[2] and x[3] and x[4] and x[5] ]

                record_ts = max([int(record[0]) for record in candles])
                record_ts_str : str = str(record_ts)
                if len(record_ts_str)==13:
                    record_ts = int(int(record_ts_str)/1000) # Convert from milli-seconds to seconds
                
                this_cutoff = record_ts  + _calc_increment(candle_size)
            else:
                this_cutoff += _calc_increment(candle_size)

        columns = ['exchange', 'symbol', 'timestamp_ms', 'open', 'high', 'low', 'close', 'volume']
        pd_all_candles = pd.DataFrame([ [ exchange.name, ticker, x[0], x[1], x[2], x[3], x[4], x[5] ] for x in all_candles], columns=columns)
        fix_column_types(pd_all_candles)
        pd_all_candles['pct_chg_on_close'] = pd_all_candles['close'].pct_change()

        rsp[ticker] = pd_all_candles

        i+=1

    return rsp

def fetch_deribit_btc_option_expiries(
    market: str = 'BTC'
) -> Dict[
    str, Union[
        Dict[str, float],
        Dict[str, Dict[str, Union[str, float]]]
    ]
]:
    exchange = ccxt.deribit()
    instruments = exchange.public_get_get_instruments({
        'currency': market,
        'kind': 'option',
        # 'expired': 'true'
    })['result']
    
    index_price = exchange.public_get_get_index_price({
        'index_name': f"{market.lower()}_usd"
    })['result']['index_price']
    index_price = float(index_price)
    
    expiry_data : Dict[str, float] = {}
    expiry_data_breakdown_by_strike : Dict[str, Dict] = {}
    for instrument in instruments:
        expiry_timestamp = int(instrument["expiration_timestamp"]) / 1000
        expiry_date = datetime.utcfromtimestamp(expiry_timestamp)

        strike = float(instrument['strike'])

        option_type = instrument['instrument_name'].split('-')[-1]  # Last part is 'C' or 'P'
        is_call = option_type == 'C'
    
        ticker = exchange.public_get_ticker({
            'instrument_name': instrument['instrument_name']
        })['result']
        
        open_interest = ticker.get("open_interest", 0)  # Open interest in BTC
        open_interest = float(open_interest)
        notional_value : float = open_interest * index_price  # Convert to USD
        
        expiry_str : str = expiry_date.strftime("%Y-%m-%d")
        if expiry_str not in expiry_data:
            expiry_data[expiry_str] = 0
        expiry_data[expiry_str] += notional_value

        if f"{expiry_str}-{strike}" not in expiry_data_breakdown_by_strike:
            expiry_data_breakdown_by_strike[f"{expiry_str}-{strike}"] = {
                'expiry' : expiry_str,
                'strike' : strike,
                'option_type': 'call' if is_call else 'put',
                'notional_value' : notional_value
            }
        else:
            expiry_data_breakdown_by_strike[f"{expiry_str}-{strike}"]['notional_value'] += notional_value
    
    sorted_expiry_data = sorted(expiry_data.items())

    return {
        'index_price' : index_price,
        'by_expiry' : sorted_expiry_data, # type: ignore Otherwise, Error: Type "dict[str, list[tuple[str, float]] | dict[str, Dict[Unknown, Unknown]]]" is not assignable to return type "Dict[str, Dict[str, float] | Dict[str, Dict[str, str | float]]]"
        'by_expiry_and_strike' : expiry_data_breakdown_by_strike
    }

def fetch_funding_rate(
    exchange,
    normalized_symbols : List[str],
    start_ts : int,
    end_ts : int,
    limit : int = 100
) -> Dict[str, pd.DataFrame]:
    results : Dict = {}
    markets = exchange.load_markets()

    funding_rate_annualized_buckets = {
        '< -15' : (float('-inf'), -15),
        '-15 - -10' : (-15, -10),
        '-10 - -8' : (-10, -8),
        '-8 - -5' : (-8, 5),
        '-5 - 0' : (-5, 0),
        '0 - 5' : (0, 5),
        '5 - 8' : (5, 8), 
        '8 - 10' : (8, 10),
        '10  - 15' : (10, 15),
        '> 15' : (15, float('inf'))
    }
    for ticker in normalized_symbols:
        market = exchange.markets[ticker] if ticker in markets else None
        this_ticker_start_ts = start_ts
        if market and market['created']:
            this_ticker_start_ts = max(this_ticker_start_ts, int(market['created']/1000))
        all_funding = []
        params = {}
        this_cutoff = this_ticker_start_ts
        while this_cutoff < end_ts:
            @retry(num_attempts=3, pause_between_retries_ms=1000, logger=None)
            def _fetch_funding_rate_history(exchange, symbol, since, limit, params):
                return exchange.fetchFundingRateHistory(symbol=symbol, since=since, limit=limit, params=params)
            funding = _fetch_funding_rate_history(exchange=exchange, symbol=ticker, since=int(this_cutoff * 1000), limit=limit, params=params)
            if funding and len(funding)>0:
                all_funding = all_funding + funding
                record_ts = max([int(entry['timestamp']) for entry in funding])
                record_ts_str : str = str(record_ts)
                if len(record_ts_str)==13:
                    record_ts = int(record_ts / 1000)
                this_cutoff = record_ts + 1
            else:
                break
        funding_history = [{
            'datetime_utc': datetime.fromtimestamp(int(entry['timestamp']/1000), tz=timezone.utc),
            'timestamp_ms': entry['timestamp'],
            'funding_rate_interval': round(entry['fundingRate'] * 100, 2),
            'funding_rate_annualized': round(entry['fundingRate'] * 100 * 3 * 365, 2),
        } for entry in all_funding]
        pd_funding_history = pd.DataFrame(funding_history)
        pd_funding_history['funding_rate_annualized_bucket'] = pd_funding_history['funding_rate_annualized'].apply(lambda x: next((k for k, (lo, hi) in funding_rate_annualized_buckets.items() if lo <= x < hi), None))
        pd_funding_history['datetime_utc'] = pd_funding_history['datetime_utc'].dt.tz_convert(None)
        results[ticker] = pd_funding_history
    return results

def build_pair_candles(
    pd_candles1 : pd.DataFrame,
    pd_candles2 : pd.DataFrame,
    left_columns_postfix : str = "_1",
    right_columns_postfix : str = "_2"
) -> pd.DataFrame:
    min_timestamp_ms1 = int(pd_candles1.iloc[0]['timestamp_ms'])
    max_timestamp_ms1 = int(pd_candles1.iloc[-1]['timestamp_ms'])
    min_timestamp_ms2 = int(pd_candles2.iloc[0]['timestamp_ms'])
    max_timestamp_ms2 = int(pd_candles2.iloc[-1]['timestamp_ms'])

    pd_candles1 = pd_candles1[(pd_candles1.timestamp_ms>=min_timestamp_ms2) & (pd_candles1.timestamp_ms<=max_timestamp_ms2) & (~pd_candles1.timestamp_ms.isna()) ]
    pd_candles2 = pd_candles2[(pd_candles2.timestamp_ms>=min_timestamp_ms1) & (pd_candles2.timestamp_ms<=max_timestamp_ms1) & (~pd_candles2.timestamp_ms.isna())]
    assert(pd_candles1.shape[0]==pd_candles2.shape[0])

    pd_candles1['timestamp_ms_gap'] = pd_candles1['timestamp_ms'] - pd_candles1['timestamp_ms'].shift(1)
    timestamp_ms_gap = pd_candles1.iloc[-1]['timestamp_ms_gap']
    
    assert(pd_candles1[~pd_candles1.timestamp_ms_gap.isna()][pd_candles1.timestamp_ms_gap!=timestamp_ms_gap].shape[0]==0)
    pd_candles1.drop(columns=['timestamp_ms_gap'], inplace=True)

    pd_candles2['timestamp_ms_gap'] = pd_candles2['timestamp_ms'] - pd_candles2['timestamp_ms'].shift(1)
    timestamp_ms_gap = pd_candles2.iloc[-1]['timestamp_ms_gap']
    assert(pd_candles2[~pd_candles2.timestamp_ms_gap.isna()][pd_candles2.timestamp_ms_gap!=timestamp_ms_gap].shape[0]==0)
    pd_candles2.drop(columns=['timestamp_ms_gap'], inplace=True)

    min_timestamp_ms1 = int(pd_candles1.iloc[0]['timestamp_ms'])
    max_timestamp_ms1 = int(pd_candles1.iloc[-1]['timestamp_ms'])
    min_timestamp_ms2 = int(pd_candles2.iloc[0]['timestamp_ms'])
    max_timestamp_ms2 = int(pd_candles2.iloc[-1]['timestamp_ms'])
    assert(min_timestamp_ms1==min_timestamp_ms2)
    assert(max_timestamp_ms1==max_timestamp_ms2)
    assert(pd_candles1.shape[0]==pd_candles2.shape[0])

    if len([ col for col in pd_candles1.columns if col[-2:]==left_columns_postfix ]) == 0:
        pd_candles1.columns = [str(col) + left_columns_postfix for col in pd_candles1.columns]

    if len([ col for col in pd_candles2.columns if col[-2:]==right_columns_postfix ]) == 0:
        pd_candles2.columns = [str(col) + right_columns_postfix for col in pd_candles2.columns]

    pd_candles1.reset_index(drop=True, inplace=True)
    pd_candles2.reset_index(drop=True, inplace=True)
    pd_candles = pd.concat([pd_candles1, pd_candles2], axis=1)
    pd_candles['timestamp_ms_gap'] = pd_candles[f'timestamp_ms{left_columns_postfix}'] - pd_candles[f'timestamp_ms{right_columns_postfix}']
    assert(pd_candles[pd_candles.timestamp_ms_gap!=0].shape[0]==0)

    pd_candles.drop(pd_candles.columns[pd_candles.columns.str.contains('unnamed',case = False)],axis = 1, inplace = True)

    return pd_candles

def get_old_ticker(
        ticker : str,
        ticker_change_map : List[Dict[str, Union[str, int]]]
    ) -> Union[str, None]:
        if not ticker_change_map:
            return None

        mapping = get_ticker_map(ticker, ticker_change_map)
        if mapping:
            return str(mapping['old_ticker'])
        
        return None

def get_ticker_map(
    ticker : str,
    ticker_change_map : List[Dict[str, Union[str, int]]]
) -> Union[None, Dict[str, Union[str, int]]]:
    if not ticker_change_map:
        return None

    for mapping in ticker_change_map:
        new_ticker = mapping['new_ticker']
        if new_ticker==ticker:
            return mapping
    
    return None

def fetch_headlines_from_rss(
    rss_feeds : Dict[str, str],
    pd_old_headlines : pd.DataFrame, # For purpose of de-duplication
    top_lines : int = 20 # For each feed, we'd only go thru top number of lines
) -> List[Dict[str, Union[str, datetime, int]]]:
    logger: logging.Logger = logging.getLogger()

    new_headlines : List[Dict[str, Union[str, datetime, int]]] = []
    for source, feed_url in rss_feeds.items():
        try:
            feed = feedparser.parse(feed_url)
            logger.info(f"{source}: {len(feed.entries)} headlines found, we're going thru top {top_lines}")

            for entry in feed.entries[:top_lines]:
                try:
                    published_dt = None
                    published = entry.get('published')
                    if published:
                        published_dt = parser.parse(entry.get('published'))
                except Exception as dateparse_err:
                    published_dt = None
                    logger.info(f"{source} {entry.title}: Date parse error {dateparse_err}")

                new_fetch_row = {
                    'source': source,
                    'title': entry.title,
                    'published_utc_dt': published_dt,
                    'published_local_dt': published_dt.astimezone() if published_dt else None,
                    'published_timestamp_ms': int(published_dt.timestamp() * 1000) if published_dt else None,
                    'created_timestamp_ms' : int(datetime.now().timestamp() * 1000),
                    'url': entry.link,
                }
                if (
                        not ((pd_old_headlines['source'] == new_fetch_row['source']) & 
                        (pd_old_headlines['title'] == new_fetch_row['title'])).any()
                ):
                    # Try not append duplicates
                    new_headlines.append(new_fetch_row)
            
        except Exception as e:
            logger.info(f"{source}: feedparser error {e}")
        
    return new_headlines



import asyncio
import os
import logging
from typing import Dict, Optional
from datetime import datetime
import time
import arrow
from dateutil import parser
import argparse
import json
from enum import Enum
import feedparser
import pandas as pd
from pprint import pformat
from tabulate import tabulate
from redis import StrictRedis

from siglab_py.util.market_data_util import fetch_headlines_from_rss
from siglab_py.util.notification_util import dispatch_notification

current_filename = os.path.basename(__file__)
current_dir : str = os.path.dirname(os.path.abspath(__file__))
parent_dir = os.path.abspath(os.path.join(current_dir, ".."))

class LogLevel(Enum):
    CRITICAL = 50
    ERROR = 40
    WARNING = 30
    INFO = 20
    DEBUG = 10
    NOTSET = 0

logging.Formatter.converter = time.gmtime
logger: logging.Logger = logging.getLogger()
log_level: int = logging.INFO
logger.setLevel(log_level)
format_str: str = '%(asctime)s %(message)s'
formatter: logging.Formatter = logging.Formatter(format_str)
sh: logging.StreamHandler = logging.StreamHandler()
sh.setLevel(log_level)
sh.setFormatter(formatter)
logger.addHandler(sh)
fh = logging.FileHandler(f"rss_headlines_provider.log")
fh.setLevel(log_level)
fh.setFormatter(formatter)     
logger.addHandler(fh)

'''
RSS feeds to fetch Financial and Geopolitical headlines:
* impact stock prices, crypto, metal and commodities prices
* macro (For example FEDs rate cut decisions, dot plot, yield curves ..etc)
* geopolitics (For example Ukraine war, US/Israel - Iran conflict Strait of Hormuz, sanctions)
* crypto exchanges, dexes hacks, security breaches, vulnerabilities may impact some crypto prices
* tokens unlocks may impact some crypto prices

Events which may move asset prices.

Usage:
    set PYTHONPATH=%PYTHONPATH%;D:\dev\siglab\siglab_py
    python headlines_provider.py --urls_list_filename headlines_rss_source.txt --focus_keywords "war, oil, trump, israel, tehran, iran, kharg, military, strike, explode, explosion, negotiate, negotiation, sanctions, nuclear, uranium" --slack_info_url https://hooks.slack.com/services/xxx --slack_critial_url https://hooks.slack.com/services/xxx --slack_alert_url https://hooks.slack.com/services/xxx

    --slack_info_url/slack_critial_url/slack_alert_url are optional

    You'd receive notification upon arrival of new headlines if:
    * slack url's provided: Lookup how to configure "Incoming WebHooks" (a slack app) under Slack's "Browse Apps"
    * focus_keywords matches headline's title

launch.json for Debugging from VSCode:
    {
        "version": "0.2.0",
        "configurations": [
            {
                "name": "Python: Current File",
                "type": "python",
                "request": "launch",
                "program": "${file}",
                "console": "integratedTerminal",
                "justMyCode": false,
                "args" : [
                        "--urls_list_filename", "headlines_rss_source.txt",
                        "--focus_keywords", "war, oil, trump, israel, tehran, iran, kharg, military, strike, explode, explosion, negotiate, negotiation, sanctions, nuclear, uranium"
                    ],
            }
        ]
    }

"headlines_rss_source.txt" takes this format:
    bloomberg|https://feeds.bloomberg.com/markets/news.rss
    bbc|http://feeds.bbci.co.uk/news/rss.xml
    cnn|http://rss.cnn.com/rss/edition.rss

"headlines_provider.py" put headlines from RSS in DataFrame, then publish to redis (optionally). Algo's or sentiment analysis can pick up from there and react to the head lines.
'''
rss_feeds = {}

param : Dict = {
    'headlines_cache_filename' : None,
    'loop_freq_ms' : 1000*60, 
    'current_filename' : current_filename,

    'notification' : {
        'footer' : None,

        # slack webhook url's for notifications
        'slack' : {
            'info' : { 'webhook_url' : None },
            'critical' : { 'webhook_url' : None },
            'alert' : { 'webhook_url' : None },
        }
    },
    
    'mds': {
        'topics': {
            'headlines': 'trading.headlines'
        },
        'redis': {
            'host': 'localhost',
            'port': 6379,
            'db': 0,
            'ttl_ms': 1000 * 60 * 15
        }
    }
}

def parse_args():
    parser = argparse.ArgumentParser() # type: ignore
    parser.add_argument("--urls_list_filename", help="File containing list of RSS url's", default=None)
    parser.add_argument("--focus_keywords", help="Comma separated list of focused keywords", default=None)
    parser.add_argument("--headlines_cache_filename", help="Export headers to csv file? Export don't filter by focus_keywords, whole data set is dumped to csv.", default='rss_headlines.csv')

    parser.add_argument("--slack_info_url", help="Slack webhook url for INFO", default=None)
    parser.add_argument("--slack_critial_url", help="Slack webhook url for CRITICAL", default=None)
    parser.add_argument("--slack_alert_url", help="Slack webhook url for ALERT", default=None)
    parser.add_argument("--enable_notification", help="Enable notification on new headline arrival? Y or N (default).", default='N')

    args = parser.parse_args()
    
    urls_list_filename : str = f"{parent_dir}\\{args.urls_list_filename}"
    print(f"Trying to read RSS url list from: {urls_list_filename}")
    with open(urls_list_filename, 'r') as urls_list:
        for line in urls_list:
            source = line.rstrip('\n').split('|')[0]
            url = line.rstrip('\n').split('|')[1]
            rss_feeds[source] = url

    if args.focus_keywords:
        param['focus_keywords'] = [ keyword.strip().lower() for keyword in args.focus_keywords.split(',') ]

    headlines_cache_filename : str = f"{parent_dir}\\{args.headlines_cache_filename}"
    param['headlines_cache_filename'] = headlines_cache_filename
    print(f"headlines cache here: {headlines_cache_filename}")

    param['notification']['slack']['info']['webhook_url'] = args.slack_info_url
    param['notification']['slack']['critical']['webhook_url'] = args.slack_critial_url
    param['notification']['slack']['alert']['webhook_url'] = args.slack_alert_url

    if args.enable_notification:
        if args.enable_notification=='Y':
            param['enable_notification'] = True
        else:
            param['enable_notification'] = False
    else:
        param['enable_notification'] = False

    param['notification']['footer'] = f"From {param['current_filename']}"

    logger.info(f"Startup args: {args}") # Dont use logger, not yet setup yet.
    logger.info(f"param: {json.dumps(param, indent=2)}")

def init_redis_client() -> StrictRedis:
    redis_client : StrictRedis = StrictRedis(
                    host = param['mds']['redis']['host'],
                    port = param['mds']['redis']['port'],
                    db = 0,
                    ssl = False
                )
    try:
        redis_client.keys()
    except ConnectionError as redis_conn_error:
        err_msg = f"Failed to connect to redis: {param['mds']['redis']['host']}, port: {param['mds']['redis']['port']}"
        logger.info(f"Failed to init redis connection. Will skip publishes to redis. {err_msg}")
        redis_client = None # type: ignore
    
    return redis_client

async def main() -> None:
    parse_args()

    notification_params : Dict[str, Any] = param['notification']

    try:        
        redis_client : Optional[StrictRedis] = init_redis_client()
    except Exception as redis_err:
        redis_client = None
        logger.info(f"Failed to connect to redis. Still run but not publishing to it. {redis_err}")

    headlines_data = []

    loop_counter : int = 0
    while True:
        try:
            start_ts_sec = time.time()

            pd_old_headlines = pd.DataFrame(columns=[ 'source', 'title', 'published_timestamp_ms', 'published_utc_dt', 'published_local_dt', 'created_timestamp_ms', 'url' ])
            if os.path.exists(param['headlines_cache_filename']):
                pd_old_headlines = pd.read_csv(param['headlines_cache_filename'], parse_dates=['published_utc_dt', 'published_local_dt'])
                pd_old_headlines.drop(pd_old_headlines.columns[pd_old_headlines.columns.str.contains('unnamed',case = False)],axis = 1, inplace = True)
                pd_old_headlines['published_timestamp_ms'] = pd_old_headlines['published_timestamp_ms'].astype(float).astype('Int64')
                if not pd_old_headlines.empty:
                    headlines_data = pd_old_headlines.to_dict('records')
                    logger.info(f"Loaded {len(headlines_data)} existing headlines from cache")
            
            new_headlines = fetch_headlines_from_rss(rss_feeds, pd_old_headlines)
            for new_headline in new_headlines:
                try:
                    logger.info(f"new row appended: {pformat(new_headline, indent=2, width=100)}")
                    new_head_line_title_words = [ x.lower().strip() for x in new_headline['title'].split(' ')]
                    if param['enable_notification'] and loop_counter>0 and (any([ keyword for keyword in param['focus_keywords'] if keyword.lower().strip() in new_head_line_title_words ])): # Don't send flood of slacks in first iteration
                        dispatch_notification(title=f"#headline [{new_headline['source']}] {new_headline['title']} ...", message=new_headline, footer=param['notification']['footer'], params=notification_params, log_level=LogLevel.CRITICAL, logger=logger)
                except Exception as notification_err:
                    pass # Just swallow
            
            num_new_headlines = len(new_headlines)
            num_old_headlines = len(headlines_data)
            headlines_data = headlines_data + new_headlines

            pd_headlines = pd.DataFrame(headlines_data)
            pd_headlines['published_timestamp_ms'] = pd_headlines['published_timestamp_ms'].fillna(0)
            pd_headlines = pd_headlines.sort_values(by=['published_timestamp_ms', 'created_timestamp_ms'], ascending=False)

            try:
                pd_headlines.to_csv(param['headlines_cache_filename'])
            except Exception as writecsv_err:
                logger.error(f"If you want to update {param['headlines_cache_filename']}, don't lock it.")

            filtered_headlines = pd_headlines[pd_headlines['title'].str.contains('|'.join(param['focus_keywords']), case=False, na=False)]

            if not filtered_headlines.empty:
                logger.info(f"{tabulate(filtered_headlines.loc[:, 'source':'created_timestamp_ms'], headers='keys', tablefmt='psql')}")

            if redis_client:
                try:
                    json_str = filtered_headlines.to_json(orient='records', date_format='iso', indent=2)

                    publish_topic = param['mds']['topics']['headlines']
                    redis_client.publish(publish_topic, json_str)
                    redis_client.setex(publish_topic, param['mds']['redis']['ttl_ms'] // 1000, json_str)
                    logger.info(f"Published filtered headlines to Redis topic {publish_topic}")

                except Exception as e:
                    logger.info(f"Failed to publish to Redis: {str(e)}")

            elapsed_ms = int((time.time() - start_ts_sec) *1000)
            logger.info(f"[loop# {loop_counter} (elapsed_ms: {elapsed_ms:,})] {pd_headlines.shape[0]} rows headlines exported to {param['headlines_cache_filename']}. # new_headlines: {num_new_headlines}, # old headlines: {num_old_headlines}")

        except Exception as loop_err:
            err_msg = f'loop error {loop_err}'
            logger.error(err_msg)
            dispatch_notification(title=f"{param['current_filename']} error. {_ticker}", message=err_msg, footer=param['notification']['footer'], params=notification_params, log_level=LogLevel.ERROR, logger=logger)
                
        finally:
            await asyncio.sleep(int(param['loop_freq_ms'] / 1000))

            loop_counter += 1

if __name__ == '__main__':
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        logger.info("Stopped by user")
    except Exception as e:
        logger.info(f"Unexpected error: {e}")


import unittest
from typing import List, Dict, Union
import time
import json
from numpy import equal
from pathlib import Path

from util.simple_str import keywords_match, classify_ticker

class SimpleStrTests(unittest.TestCase):
    
    def test_keywords_match(self):
        '''
        Folder structure:
            \ siglab
                \ siglab_py	<-- python project root
                    \ sigab_py
                        __init__.py
                        \ util
                            __init__.py
                            market_data_util.py
                        \ tests
                            \ unit
                                __init__.py
                                analytic_util_tests.py <-- Tests here
                            
                \ siglab_rs <-- Rust project root
                \ data	 <-- Data files here!
        '''
        data_dir = Path(__file__).parent.parent.parent.parent / "data"
        with open(f"{data_dir}\\market_impact_keywords.json", 'r') as f:
            keywords_cache = json.load(f)
        fuzzy_threshold = 80
        # Note: Israel -> Isreel
        example = "Iran and Isreel are in tense negotiations. Oil prices surged after a missile strike near the Strait of Hormuz. The Fed might pause rates cuts."
        matches = keywords_match(
            sentence=example, 
            keywords_cache=keywords_cache, 
            fuzzy=True, 
            fuzzy_threshold=fuzzy_threshold
        )
        print(json.dumps(matches, indent=2))

        assert('num_matches' in matches)
        assert('word_count' in matches)
        assert('matches_percent' in matches)
        assert(matches['num_matches']>=0)
        assert(matches['word_count']>0)
        assert(matches['matches_percent']<=100)
        assert('nouns' in matches)
        assert('actions' in matches)
        assert('adjectives' in matches)

        assert(matches['matches_percent']>40)

        assert(matches['nouns']['countries']==['israel', 'iran'])
        assert(matches['nouns']['financial_institutions']==['fed'])
        assert(matches['nouns']['infrastructure']==['strait of hormuz'])
        assert(matches['nouns']['commodities']==['oil'])

        assert(matches['actions']['diplomacy']==['negotiation'])
        assert(matches['actions']['attack']==['strike', 'missile'])
        assert(matches['actions']['economic_policy']==['cut'])
        assert(matches['actions']['market_movement']==['surge'])

        assert(matches['adjectives']['geopolitical']==['tense'])

    def test_classify_ticker(self):
        cases = [
            # Crypto spot
            ("BTC/USDT",                "spot"),
            ("ETH/BTC",                 "spot"),
            ("1000PEPE/USDT",           "spot"),
            ("NVDA/USDC",               "spot"),
            ("EUR/JPY",                 "spot"),
            ("AUD/NZD",                 "spot"),

            # Crypto perpetual
            ("BTC/USDT:USDT",           "crypto.perpetual"),
            ("NVDA/USDC:USDC",          "crypto.perpetual"),
            ("XRP/USDT:1000XRP",        "crypto.perpetual"),

            # Crypto dated future & option
            ("BTC-26MAR26",             "crypto.dated_future"),
            ("BTC-22MAR26-68500-C",     "crypto.option"),
            ("ETH-25DEC25",             "crypto.dated_future"),
            ("BTC-USD-251226-100000-C", "crypto.option"),

            # TradFi / IBKR style
            ("AAPL",                    "tradfi.stock"),
            ("TSLA",                    "tradfi.stock"),
            ("IBM-STK-SMART-USD",       "tradfi.stock"),
            ("ESU6",                    "tradfi.future"),       # Sep 2026
            ("NQZ5",                    "tradfi.future"),
            ("CLM6",                    "tradfi.future"),
            ("AAPL-OPT-20260320-250-C-SMART-100-USD", "tradfi.option"),
            ("SPY-OPT-20260618-550-P-CBOE", "tradfi.option"),
            ("ES-FOP-20260320-5000-C-CME",  "tradfi.option"),

            # Edge / invalid
            ("BTCUSDT",                 "tradfi.stock"),
            ("BTC-USDT",                "unknown"),
            ("@ES",                     "unknown"),       # continuous future
            ("SPY",                     "tradfi.stock"),  # ETF
        ]

        for ticker, expected in cases:
            result = classify_ticker(ticker)
            assert result == expected, \
                f"Failed: {ticker!r} → got {result!r}, expected {expected!r}"


'''
https://github.com/FutunnOpen/py-futu-api/blob/master/README.md
https://www.futuhk.com/en/support/categories/909?global_content=%7B%22promote_id%22%3A13765%2C%22sub_promote_id%22%3A10%7D

Fees: https://www.futuhk.com/en/commissionnew#crypto

Subscribe L2 data: https://openapi.futunn.com/futu-api-doc/en/intro/authority.html

Investor Protection: https://www.futuhk.com/en

Margin Trading: 
    https://www.futunn.com/en/learn/detail-what-is-margin-trading-62010-220679271
    Margin rate 6.8% annual https://www.futuhk.com/en/support/topic2_417

Download Futu OpenD
    https://www.futuhk.com/en/support/topic1_464?global_content=%7B%22promote_id%22%3A13765%2C%22sub_promote_id%22%3A10%7D

    If you run the installer version "Futu_OpenD-GUI_9.0.5008_Windows.exe", it'd be installed under C-Drive:
        C:\\Users\\xxx\\AppData\\Roaming\\Futu_OpenD\\Futu_OpenD.exe
    Unfortunately, log folder also under C-drive as a result, and they are big.

    For command line version: https://openapi.futunn.com/futu-api-doc/opend/opend-cmd.html
        Binary under downloaded package (You can put it under for example D-drive): 
            ...\Futu_OpenD_9.4.5408_Windows\Futu_OpenD_9.4.5408_Windows
            
        Put a batch file "start_futu_opend.bat", if login_pwd include special characters, enclose pwd with double quotes:
            FutuOpenD -login_account=1234567 -login_pwd="... Your Secret here ..."
    Config file is "FutuOpenD.xml", you can adjust logging verbosity here.

Architecture: https://openapi.futunn.com/futu-api-doc/en/intro/intro.html

python -mpip install futu-api

Whatsapp support: Moomoo OpenAPI 1群

API
    Fetches:
        market status https://openapi.futunn.com/futu-api-doc/en/quote/get-global-state.html
        stock basic info https://openapi.futunn.com/futu-api-doc/en/quote/get-static-info.html
        historical candles https://openapi.futunn.com/futu-api-doc/en/quote/request-history-kline.html
        realtime candles https://openapi.futunn.com/futu-api-doc/en/quote/get-kl.html
        orderbook https://openapi.futunn.com/futu-api-doc/en/quote/get-order-book.html
        real time quote https://openapi.futunn.com/futu-api-doc/en/quote/get-stock-quote.html
        open orders https://openapi.futunn.com/futu-api-doc/en/trade/get-order-list.html
        historical orders https://openapi.futunn.com/futu-api-doc/en/trade/get-history-order-list.html
        positions https://openapi.futunn.com/futu-api-doc/en/trade/get-position-list.html
        balances https://openapi.futunn.com/futu-api-doc/en/trade/get-funds.html

    Trading:
        create order https://openapi.futunn.com/futu-api-doc/en/trade/place-order.html
        amend order https://openapi.futunn.com/futu-api-doc/en/trade/modify-order.html

'''
from datetime import datetime, timedelta
import pytz
from typing import List, Dict, Union, Any, NoReturn
import pandas as pd

import futu as ft
from futu import *

from siglab_py.exchanges.any_exchange import AnyExchange

class Futubull(AnyExchange):
    def __init__(self, *args: Dict[str, Any]) -> None:
        super().__init__(*args)
        
        self.name = 'futubull'

        params : Dict[str, Any] = args[0] # Type "object" is not assignable to declared type "Dict[Unknown, Unknown]"
        self.host_addr = params['daemon']['host']
        self.port = params['daemon']['port']
        self.market = params['market']
        self.security_type = params['security_type']
        self.quote_ctx = OpenQuoteContext(host=self.host_addr, port=self.port)
        self.trd_ctx = OpenSecTradeContext(
                filter_trdmarket=params['trdmarket'], 
                host=self.host_addr, port=self.port, 
                security_firm=params['security_firm']
            )
        self.markets : Dict = {}

    def load_markets(self, reload=False, params={}):  # type: ignore
        '''
        Mimic CCXT load_markets https://github.com/ccxt/ccxt/blob/master/python/ccxt/base/exchange.py

        Examplem,
            {
                ... more pairs ...
                'ETH/USDC:USDC': {
                    'id': 'ETH-USDC-SWAP',
                    'lowercaseId': None,
                    'symbol': 'ETH/USDC:USDC',
                    'base': 'ETH',
                    'quote': 'USDC',
                    'settle': 'USDC',
                    'baseId': 'ETH',
                    'quoteId': 'USDC',
                    'settleId': 'USDC',
                    'type': 'swap',
                    'spot': False,
                    'margin': False,
                    'swap': True,
                    'future': False,
                    'option': False,
                    'index': None,
                    'active': True,
                    'contract': True,
                    'linear': True,
                    'inverse': False,
                    'subType': 'linear',
                    'taker': 0.0005,
                    'maker': 0.0002,
                    'contractSize': 0.001,
                    'expiry': None,
                    'expiryDatetime': None,
                    'strike': None,
                    'optionType': None,
                    'precision': {'amount': 1.0, 'price': 0.01, 'cost': None, 'base': None, 'quote': None},
                    'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 1.0, 'max': None}, 'price': {'min': None, 'max': None}, 'cost': {'min': None, 'max': None}},
                    'marginModes': {'cross': None, 'isolated': None},
                    'created': 1666076197702,
                    'info' : {
                        ... raw exchange response here ...
                    }
                },
                ... more pairs ...
            }

        gateway.py will call:
            1. multiplier = market['contractSize'] (Note this is contract size for futures, not same as 'Lot Size')
            2. price_to_precision -> market['precision']['price']
            3. amount_to_precision  -> market['precision']['amount']    <-- This is 'Lot Size'
        '''
        
        ret, data = self.quote_ctx.get_stock_basicinfo(self.market, self.security_type)
        if ret == RET_OK:
            for index, row in data.iterrows(): # type: ignore
                symbol = row['code']

                name = row['name']
                stock_type = row['stock_type']
                stock_child_type = row['stock_child_type']
                stock_owner = row['stock_owner']
                option_type = row['option_type']
                strike_time = row['strike_time']
                strike_price = row['strike_price']
                suspension = row['suspension']
                stock_id = row['stock_id']
                listing_date = row['listing_date']
                delisting = bool(row['delisting'])
                main_contract = bool(row['main_contract'])
                last_trade_time = row['last_trade_time']
                exchange_type = row['exchange_type']
                lot_size = row['lot_size']
                
                # No additional information from 'get_stock_basicinfo'
                # ret, detail = self.quote_ctx.get_stock_basicinfo(self.market, self.security_type, [ symbol ])

                info : Dict = {
                    'symbol' : symbol,
                    'name' : name,
                    'stock_type' : stock_type,
                    'stock_child_type' : stock_child_type,
                    'stock_owner' : stock_owner,
                    'option_type' : option_type,
                    'strike_time' : strike_time,
                    'strike_price' : strike_price,
                    'suspension' : suspension,
                    'stock_id' : stock_id,
                    'listing_date' : listing_date,
                    'delisting' : delisting,
                    'main_contract' : main_contract,
                    'last_trade_time' : last_trade_time,
                    'exchange_type' : exchange_type,
                    'lot_size' : lot_size
                }
                self.markets[symbol] = {
                    'symbol' : symbol,
                    'id' : symbol,

                    'base': None,
                    'quote': None,
                    'settle': None,
                    'baseId': None,
                    'quoteId': None,
                    'settleId': None,

                    'type': self.security_type,
                    'spot': False,
                    'margin': False,
                    'swap': False,
                    'future': False,
                    'option': False,
                    'index': None,
                    'active': not delisting,
                    'contract': False,
                    'linear': False,
                    'inverse': False,
                    'subType': False,
                    'taker': 0,
                    'maker': 0,
                    'contractSize': 0,
                    'expiry': strike_time,
                    'expiryDatetime': strike_time,
                    'strike': strike_price,
                    'optionType': option_type,

                    'precision': {
                        'amount': lot_size, 
                        'price': 0.01, 
                        'cost': None, 
                        'base': None, 
                        'quote': None
                    },

                    'limits': {
                        'leverage': {'min': 1, 'max': 5}, 
                        'amount': {'min': 1.0, 'max': None}, 
                        'price': {'min': None, 'max': None}, 
                        'cost': {'min': None, 'max': None}
                    },
                    'marginModes': {'cross': None, 'isolated': None},

                    'info' : info
                }
        return self.markets

    '''
    With CCXT, 'params' is used to parse exchange specific parameters. Not actually used with Futu.

    Get Historical Candlesticks max_count: In API doc it advises<1000, but when I tested, it's 246 (1 year).
    To fetch more than this, use market_data_util.fetch_candles (Sliding window implemented).
    Under the hood, it'd use implementation below.

    CCXT fetch_ohlcv response format is:
        [
            [1704038400000, 42469.9, 42723.1, 42442.3, 42609.8, 3624.59]
            [1704042000000, 42609.8, 42672.4, 42473.9, 42630.6, 2829.69]
            [1704045600000, 42630.6, 42742.9, 42542.9, 42673.8, 2097.12]
            [1704049200000, 42673.8, 42705.0, 42593.7, 42627.7, 2055.54]
            [1704052800000, 42627.6, 42693.0, 42511.1, 42565.5, 2061.97]
            ... more candles ...
        ]
    
    Fields are: timestamp (in ms), open, high, low, close, volume.
    '''
    def fetch_ohlcv(self, symbol: str, timeframe : str ='1h', since: Union[int, None] = None, limit: int = 100, params={}) -> Union[List[List[Union[int, float]]], None]:  # type: ignore
        if not symbol or not since:
            return None
        
        ktype=KLType.K_DAY
        if timeframe=="1m":
            ktype=KLType.K_1M
        elif timeframe=="1h":
            ktype=KLType.K_60M
        elif timeframe=="1h":
            ktype=KLType.K_DAY

        candles = []
        dt_start = datetime.fromtimestamp(int(since/1000))
        s_start : str = dt_start.strftime("%Y-%m-%d")
        ret, data, page_req_key = self.quote_ctx.request_history_kline(code=symbol, start=s_start, ktype=ktype, max_count=limit)
        if ret == RET_OK:
            for index, row in data.iterrows(): # type: ignore

                '''
                From doc:
                    Format: yyyy-MM-dd HH:mm:ss
                    The default of HK stock market and A-share market is Beijing time, 
                    while that of US stock market is US Eastern time.
                '''
                time_key = row['time_key']
                dt = datetime.strptime(time_key, "%Y-%m-%d %H:%M:%S")
                
                if self.market in [ Market.HK, Market.SH, Market.SZ ]:
                    tz = pytz.timezone('Asia/Shanghai')
                    dt = tz.localize(dt)
                elif self.market == Market.US:
                    tz = pytz.timezone('US/Eastern')
                    dt = tz.localize(dt)
                elif self.market == Market.CA:
                    tz = pytz.timezone('US/Eastern')
                    dt = tz.localize(dt)
                elif self.market == Market.AU:
                    tz = pytz.timezone('Australia/Sydney')
                    dt = tz.localize(dt)
                else:
                    # @todo:  HK SH SZ US AU CA FX
                    raise ValueError(f"Unsupported market {self.market}")

                timestamp_ms = int(dt.timestamp() * 1000)
                open = row['open']
                high = row['high']
                low = row['low']
                close = row['close']
                volume = row['volume']
                candles.append(
                    [ timestamp_ms, open, high, low, close, volume ]
                )
        return candles

    def fetch_candles(
        self,
        start_ts,
        end_ts,
        symbols,
        candle_size
    ) -> Dict[str, Union[pd.DataFrame, None]]:
        exchange_candles : Dict[str, Union[pd.DataFrame, None]] = {}

        for symbol in symbols:
            pd_candles = self._fetch_candles(symbol=symbol, start_ts=start_ts, end_ts=end_ts, candle_size=candle_size)
            pd_candles['exchange'] = self.name
            exchange_candles[symbol] = pd_candles

        return exchange_candles

    def _fetch_candles(
        self,
        symbol : str,
        start_ts : int,
        end_ts : int,
        candle_size : str = '1d',
        num_candles_limit : int = 100
    ):
        def _fetch_ohlcv(symbol, timeframe, since, limit, params) -> Union[List[List[Union[int, float]]], None]:
            one_timeframe = f"1{timeframe[-1]}"
            candles = self.fetch_ohlcv(symbol=symbol, timeframe=one_timeframe, since=since, limit=limit, params=params)
            if candles and len(candles)>0:
                candles.sort(key=lambda x : x[0], reverse=False)

            return candles

        all_candles = []
        params = {}
        this_cutoff = start_ts
        while this_cutoff<=end_ts:
            candles = _fetch_ohlcv(symbol=symbol, timeframe=candle_size, since=int(this_cutoff * 1000), limit=num_candles_limit, params=params)
            if candles and len(candles)>0:
                all_candles = all_candles + [[ int(x[0]), float(x[1]), float(x[2]), float(x[3]), float(x[4]), float(x[5]) ] for x in candles if x[1] and x[2] and x[3] and x[4] and x[5] ]

                record_ts = max([int(record[0]) for record in candles])
                record_ts_str : str = str(record_ts)
                if len(record_ts_str)==13:
                    record_ts = int(int(record_ts_str)/1000) # Convert from milli-seconds to seconds

                if this_cutoff==record_ts+1:
                    break
                else:
                    this_cutoff = record_ts  + 1
        columns = ['exchange', 'symbol', 'timestamp_ms', 'open', 'high', 'low', 'close', 'volume']
        pd_all_candles = pd.DataFrame([ [ "futubull", symbol, x[0], x[1], x[2], x[3], x[4], x[5] ] for x in all_candles], columns=columns)
        # fix_column_types(pd_all_candles)
        pd_all_candles['pct_chg_on_close'] = pd_all_candles['close'].pct_change()
        return pd_all_candles

if __name__ == '__main__':
    params : Dict = {
        'trdmarket' : TrdMarket.HK,
        'security_firm' : SecurityFirm.FUTUSECURITIES,
        'market' : Market.HK, 
        'security_type' : SecurityType.STOCK,
        'daemon' : {
            'host' : '127.0.0.1',
            'port' : 11111
        }
    }
    exchange = Futubull(params)
    markets = exchange.load_markets()

    dt_end = datetime.today()
    dt_start : datetime = dt_end - timedelta(days=365*3)
    timestamp_ms : int = int(dt_start.timestamp() * 1000)
    candles = exchange.fetch_ohlcv(symbol="HK.00700", timeframe="1h", since=timestamp_ms, limit=1000)

    '''
    Examples from Futu https://openapi.futunn.com/futu-api-doc/en

    '''
    quote_ctx = OpenQuoteContext(host='127.0.0.1', port=11111)

    print(quote_ctx.get_global_state())

    ret, data = quote_ctx.get_stock_basicinfo(Market.HK, SecurityType.STOCK)
    if ret == RET_OK:
        print(data)
    else:
        print('error:', data)
    print('******************************************')
    ret, data = quote_ctx.get_stock_basicinfo(Market.HK, SecurityType.STOCK, ['HK.06998', 'HK.00700'])
    if ret == RET_OK:
        assert isinstance(data, pd.DataFrame)
        print(data)
        print(data['name'][0])
        print(data['name'].values.tolist())
    else:
        print('error:', data)

    '''
    KLType: https://openapi.futunn.com/futu-api-doc/en/quote/quote.html#66
    '''
    ret, data, page_req_key = quote_ctx.request_history_kline('HK.00700', start='2024-01-01', end='2025-03-07', ktype=KLType.K_60M, max_count=100)
    if ret == RET_OK:
        assert isinstance(data, pd.DataFrame)
        print(data)
        print(data['code'][0])
        print(data['close'].values.tolist())
    else:
        print('error:', data)
    while page_req_key != None:
        print('*************************************')
        ret, data, page_req_key = quote_ctx.request_history_kline('HK.00700', start='2019-09-11', end='2019-09-18', max_count=5,page_req_key=page_req_key) # Request the page after turning data
        if ret == RET_OK:
            print(data)
        else:
            print('error:', data)

    # For any real time data, you need subscribe https://openapi.futunn.com/futu-api-doc/en/intro/authority.html#5331
    ret_sub, err_message = quote_ctx.subscribe(['HK.00700'], [SubType.K_DAY], subscribe_push=False)
    if ret_sub == RET_OK:
        ret, data = quote_ctx.get_cur_kline('HK.00700', 2, SubType.K_DAY, AuType.QFQ)
        if ret == RET_OK:
            assert isinstance(data, pd.DataFrame)
            print(data)
            print(data['turnover_rate'][0])
            print(data['turnover_rate'].values.tolist())
        else:
            print('error:', data)
    else:
        # (-1, '无权限订阅HK.00005的行情，请检查香港市场股票行情权限')
        print('subscription failed', err_message)

    ret_sub, err_message = quote_ctx.subscribe(['HK.00700'], [SubType.QUOTE], subscribe_push=False)
    if ret_sub == RET_OK:
        ret, data = quote_ctx.get_stock_quote(['HK.00700'])
        if ret == RET_OK:
            assert isinstance(data, pd.DataFrame)
            print(data)
            print(data['code'][0])
            print(data['code'].values.tolist())
        else:
            print('error:', data)
    else:
        # '无权限订阅HK.00700的行情，请检查香港市场股票行情权限'
        print('subscription failed', err_message)

    ret_sub = quote_ctx.subscribe(['HK.00700'], [SubType.ORDER_BOOK], subscribe_push=False)[0]
    if ret_sub == RET_OK:
        ret, data = quote_ctx.get_order_book('HK.00700', num=3)
        if ret == RET_OK:
            print(data)
        else:
            print('error:', data)
    else:
        print('subscription failed')

    trd_ctx = OpenSecTradeContext(filter_trdmarket=TrdMarket.HK, host='127.0.0.1', port=11111, security_firm=SecurityFirm.FUTUSECURITIES)
    ret, data = trd_ctx.accinfo_query()
    if ret == RET_OK:
        assert isinstance(data, pd.DataFrame)
        print(data)
        print(data['power'][0])
        print(data['power'].values.tolist())
    else:
        print('accinfo_query error: ', data)

    ret, data = trd_ctx.order_list_query()
    if ret == RET_OK:
        assert isinstance(data, pd.DataFrame)
        print(data)
        if data.shape[0] > 0:
            print(data['order_id'][0])
            print(data['order_id'].values.tolist())
    else:
        print('order_list_query error: ', data)

    ret, data = trd_ctx.history_order_list_query()
    if ret == RET_OK:
        assert isinstance(data, pd.DataFrame)
        print(data)
        if data.shape[0] > 0:
            print(data['order_id'][0])
            print(data['order_id'].values.tolist())
    else:
        print('history_order_list_query error: ', data)

    ret, data = trd_ctx.position_list_query()
    if ret == RET_OK:
        assert isinstance(data, pd.DataFrame)
        print(data)
        if data.shape[0] > 0:
            print(data['stock_name'][0])
            print(data['stock_name'].values.tolist())
    else:
        print('position_list_query error: ', data)

    trd_ctx.close()